@@ -16,48 +16,39 @@
 
 use crate::{
     ahp::{matrices, prover::ProverConstraintSystem, verifier, AHPError, CircuitInfo},
+    marlin::MarlinMode,
     String,
     ToString,
     Vec,
 };
-use snarkvm_algorithms::{cfg_iter_mut, fft::EvaluationDomain};
-use snarkvm_fields::{batch_inversion, Field, PrimeField};
+use snarkvm_algorithms::fft::EvaluationDomain;
+use snarkvm_fields::{Field, PrimeField};
 use snarkvm_r1cs::errors::SynthesisError;
 
 use snarkvm_polycommit::{LCTerm, LabeledPolynomial, LinearCombination};
 
 use core::{borrow::Borrow, marker::PhantomData};
 
-#[cfg(feature = "parallel")]
-use rayon::prelude::*;
-
 /// The algebraic holographic proof defined in [CHMMVW19](https://eprint.iacr.org/2019/1047).
 /// Currently, this AHP only supports inputs of size one
 /// less than a power of 2 (i.e., of the form 2^n - 1).
-pub struct AHPForR1CS<F: Field> {
+pub struct AHPForR1CS<F: Field, MM: MarlinMode> {
     field: PhantomData<F>,
+    mode: PhantomData<MM>,
 }
 
-impl<F: PrimeField> AHPForR1CS<F> {
+impl<F: PrimeField, MM: MarlinMode> AHPForR1CS<F, MM> {
     /// The labels for the polynomials output by the AHP indexer.
     #[rustfmt::skip]
-    pub const INDEXER_POLYNOMIALS: [&'static str; 12] = [
-        // Polynomials for A
-        "a_row", "a_col", "a_val", "a_row_col",
-        // Polynomials for B
-        "b_row", "b_col", "b_val", "b_row_col",
-        // Polynomials for C
-        "c_row", "c_col", "c_val", "c_row_col",
+    pub const INDEXER_POLYNOMIALS: [&'static str; 6] = [
+        // Polynomials for M
+        "row", "col", "a_val", "b_val", "c_val", "row_col",
     ];
     /// The labels for the polynomials output and vanishing polynomials by the AHP indexer.
     #[rustfmt::skip]
-    pub const INDEXER_POLYNOMIALS_WITH_VANISHING: [&'static str; 14] = [
-        // Polynomials for A
-        "a_row", "a_col", "a_val", "a_row_col",
-        // Polynomials for B
-        "b_row", "b_col", "b_val", "b_row_col",
-        // Polynomials for C
-        "c_row", "c_col", "c_val", "c_row_col",
+    pub const INDEXER_POLYNOMIALS_WITH_VANISHING: [&'static str; 8] = [
+        // Polynomials for M
+        "row", "col", "a_val", "b_val", "c_val", "row_col",
         // Vanishing polynomials
         "vanishing_poly_h", "vanishing_poly_k"
     ];
@@ -66,24 +57,40 @@ impl<F: PrimeField> AHPForR1CS<F> {
     pub const LC_WITH_ZERO_EVAL: [&'static str; 2] = ["inner_sumcheck", "outer_sumcheck"];
     /// The labels for the polynomials output by the AHP prover.
     #[rustfmt::skip]
-    pub const PROVER_POLYNOMIALS: [&'static str; 9] = [
+    pub const PROVER_POLYNOMIALS_WITHOUT_ZK: [&'static str; 8] = [
+        // First sumcheck
+        "w", "z_a", "z_b", "t", "g_1", "h_1",
+        // Second sumcheck
+        "g_2", "h_2",
+    ];
+    /// The labels for the polynomials output by the AHP prover.
+    #[rustfmt::skip]
+    pub const PROVER_POLYNOMIALS_WITH_ZK: [&'static str; 9] = [
         // First sumcheck
         "w", "z_a", "z_b", "mask_poly", "t", "g_1", "h_1",
         // Second sumcheck
         "g_2", "h_2",
     ];
 
-    pub(crate) fn polynomial_labels() -> impl Iterator<Item = String> {
-        Self::INDEXER_POLYNOMIALS
-            .iter()
-            .chain(&Self::PROVER_POLYNOMIALS)
-            .map(|s| s.to_string())
+    pub(crate) fn indexer_polynomials() -> impl Iterator<Item = &'static str> {
+        if MM::RECURSION {
+            Self::INDEXER_POLYNOMIALS_WITH_VANISHING.as_ref().iter().copied()
+        } else {
+            Self::INDEXER_POLYNOMIALS.as_ref().iter().copied()
+        }
     }
 
-    pub(crate) fn polynomial_labels_with_vanishing() -> impl Iterator<Item = String> {
-        Self::INDEXER_POLYNOMIALS_WITH_VANISHING
-            .iter()
-            .chain(&Self::PROVER_POLYNOMIALS)
+    pub(crate) fn prover_polynomials() -> impl Iterator<Item = &'static str> {
+        if MM::ZK {
+            Self::PROVER_POLYNOMIALS_WITH_ZK.as_ref().iter().copied()
+        } else {
+            Self::PROVER_POLYNOMIALS_WITHOUT_ZK.as_ref().iter().copied()
+        }
+    }
+
+    pub(crate) fn polynomial_labels() -> impl Iterator<Item = String> {
+        Self::indexer_polynomials()
+            .chain(Self::prover_polynomials())
             .map(|s| s.to_string())
     }
 
@@ -108,12 +115,18 @@ impl<F: PrimeField> AHPForR1CS<F> {
             .ok_or(SynthesisError::PolynomialDegreeTooLarge)?;
         let domain_k_size = EvaluationDomain::<F>::compute_size_of_domain(num_non_zero)
             .ok_or(SynthesisError::PolynomialDegreeTooLarge)?;
+
         Ok(*[
             2 * domain_h_size + zk_bound - 2,
-            3 * domain_h_size + 2 * zk_bound - 3, //  mask_poly
+            if MM::ZK {
+                3 * domain_h_size + 2 * zk_bound - 3
+            } else {
+                0
+            }, //  mask_poly
             domain_h_size,
             domain_h_size,
-            3 * domain_k_size - 3,
+            domain_k_size - 1,
+            domain_k_size, //  due to vanishing polynomial; for convenience, we increase the number by one regardless of the mode.
         ]
         .iter()
         .max()
@@ -134,12 +147,12 @@ impl<F: PrimeField> AHPForR1CS<F> {
     }
 
     /// Construct the linear combinations that are checked by the AHP.
+    /// Public input should be unformatted.
     #[allow(non_snake_case)]
     pub fn construct_linear_combinations<E: EvaluationsProvider<F>>(
         public_input: &[F],
         evals: &E,
-        state: &verifier::VerifierState<F>,
-        with_vanishing: bool,
+        state: &verifier::VerifierState<F, MM>,
     ) -> Result<Vec<LinearCombination<F>>, AHPError> {
         let domain_h = state.domain_h;
         let domain_k = state.domain_k;
@@ -184,21 +197,19 @@ impl<F: PrimeField> AHPForR1CS<F> {
             .fold(F::zero(), |x, y| x + y);
 
         #[rustfmt::skip]
-        let outer_sumcheck = LinearCombination::new(
-            "outer_sumcheck",
-            vec![
-                (F::one(), "mask_poly".into()),
-
-                (r_alpha_at_beta * (eta_a + (eta_c * z_b_at_beta)), "z_a".into()),
-                (r_alpha_at_beta * eta_b * z_b_at_beta, LCTerm::One),
-
-                (-t_at_beta * v_X_at_beta, "w".into()),
-                (-t_at_beta * x_at_beta, LCTerm::One),
-
-                (-v_H_at_beta, "h_1".into()),
-                (-beta * g_1_at_beta, LCTerm::One),
-            ],
-        );
+        let outer_sumcheck = {
+            let mut lc_terms = vec![];
+            if MM::ZK {
+                lc_terms.push((F::one(), "mask_poly".into()));
+            }
+            lc_terms.push((r_alpha_at_beta * (eta_a + (eta_c * z_b_at_beta)), "z_a".into()));
+            lc_terms.push((r_alpha_at_beta * eta_b * z_b_at_beta, LCTerm::One));
+            lc_terms.push((-t_at_beta * v_X_at_beta, "w".into()));
+            lc_terms.push((-t_at_beta * x_at_beta, LCTerm::One));
+            lc_terms.push((-v_H_at_beta, "h_1".into()));
+            lc_terms.push((-beta * g_1_at_beta, LCTerm::One));
+            LinearCombination::new("outer_sumcheck", lc_terms)
+        };
         debug_assert!(evals.get_lc_eval(&outer_sumcheck, beta)?.is_zero());
 
         linear_combinations.push(z_b);
@@ -210,58 +221,31 @@ impl<F: PrimeField> AHPForR1CS<F> {
         let beta_alpha = beta * alpha;
         let g_2 = LinearCombination::new("g_2", vec![(F::one(), "g_2")]);
 
-        let a_denom = LinearCombination::new("a_denom", vec![
-            (beta_alpha, LCTerm::One),
-            (-alpha, "a_row".into()),
-            (-beta, "a_col".into()),
-            (F::one(), "a_row_col".into()),
-        ]);
-
-        let b_denom = LinearCombination::new("b_denom", vec![
-            (beta_alpha, LCTerm::One),
-            (-alpha, "b_row".into()),
-            (-beta, "b_col".into()),
-            (F::one(), "b_row_col".into()),
-        ]);
-
-        let c_denom = LinearCombination::new("c_denom", vec![
-            (beta_alpha, LCTerm::One),
-            (-alpha, "c_row".into()),
-            (-beta, "c_col".into()),
-            (F::one(), "c_row_col".into()),
-        ]);
-
-        let a_denom_at_gamma = evals.get_lc_eval(&a_denom, gamma)?;
-        let b_denom_at_gamma = evals.get_lc_eval(&b_denom, gamma)?;
-        let c_denom_at_gamma = evals.get_lc_eval(&c_denom, gamma)?;
         let g_2_at_gamma = evals.get_lc_eval(&g_2, gamma)?;
 
         let v_K_at_gamma = domain_k.evaluate_vanishing_polynomial(gamma);
 
-        let mut a = LinearCombination::new("a_poly", vec![
-            (eta_a * b_denom_at_gamma * c_denom_at_gamma, "a_val"),
-            (eta_b * a_denom_at_gamma * c_denom_at_gamma, "b_val"),
-            (eta_c * b_denom_at_gamma * a_denom_at_gamma, "c_val"),
-        ]);
-
+        let mut a = LinearCombination::new("a_poly", vec![(eta_a, "a_val"), (eta_b, "b_val"), (eta_c, "c_val")]);
         a *= v_H_at_alpha * v_H_at_beta;
-        let b_at_gamma = a_denom_at_gamma * b_denom_at_gamma * c_denom_at_gamma;
-        let b_expr_at_gamma = b_at_gamma * (gamma * g_2_at_gamma + (t_at_beta / &k_size));
 
-        a -= &LinearCombination::new("b_expr", vec![(b_expr_at_gamma, LCTerm::One)]);
-        a -= &LinearCombination::new("h_2", vec![(v_K_at_gamma, "h_2")]);
+        let mut b = LinearCombination::new("denom", vec![
+            (beta_alpha, LCTerm::One),
+            (-alpha, "row".into()),
+            (-beta, "col".into()),
+            (F::one(), "row_col".into()),
+        ]);
+        b *= gamma * g_2_at_gamma + (t_at_beta / k_size);
 
-        a.label = "inner_sumcheck".into();
-        let inner_sumcheck = a;
+        let mut inner_sumcheck = a;
+        inner_sumcheck -= &b;
+        inner_sumcheck -= &LinearCombination::new("h_2", vec![(v_K_at_gamma, "h_2")]);
+        inner_sumcheck.label = "inner_sumcheck".into();
         debug_assert!(evals.get_lc_eval(&inner_sumcheck, gamma)?.is_zero());
 
         linear_combinations.push(g_2);
-        linear_combinations.push(a_denom);
-        linear_combinations.push(b_denom);
-        linear_combinations.push(c_denom);
         linear_combinations.push(inner_sumcheck);
 
-        if with_vanishing {
+        if MM::RECURSION {
             let vanishing_poly_h_alpha =
                 LinearCombination::new("vanishing_poly_h_alpha", vec![(F::one(), "vanishing_poly_h")]);
             let vanishing_poly_h_beta =
@@ -335,7 +319,7 @@ pub trait UnnormalizedBivariateLagrangePoly<F: PrimeField> {
 impl<F: PrimeField> UnnormalizedBivariateLagrangePoly<F> for EvaluationDomain<F> {
     fn eval_unnormalized_bivariate_lagrange_poly(&self, x: F, y: F) -> F {
         if x != y {
-            (self.evaluate_vanishing_polynomial(x) - self.evaluate_vanishing_polynomial(y)) / &(x - y)
+            (self.evaluate_vanishing_polynomial(x) - self.evaluate_vanishing_polynomial(y)) / (x - y)
         } else {
             self.size_as_field_element * x.pow(&[(self.size() - 1) as u64])
         }
@@ -344,9 +328,7 @@ impl<F: PrimeField> UnnormalizedBivariateLagrangePoly<F> for EvaluationDomain<F>
     fn batch_eval_unnormalized_bivariate_lagrange_poly_with_diff_inputs(&self, x: F) -> Vec<F> {
         let vanish_x = self.evaluate_vanishing_polynomial(x);
         let mut inverses: Vec<F> = self.elements().map(|y| x - y).collect();
-        batch_inversion(&mut inverses);
-
-        cfg_iter_mut!(inverses).for_each(|denominator| *denominator *= &vanish_x);
+        snarkvm_fields::batch_inversion_and_mul(&mut inverses, &vanish_x);
         inverses
     }
 
