@@ -14,7 +14,7 @@
 // You should have received a copy of the GNU General Public License
 // along with the snarkVM library. If not, see <https://www.gnu.org/licenses/>.
 
-use crate::ahp::matrices::make_matrices_square;
+use crate::{ahp::matrices::make_matrices_square, Vec};
 use snarkvm_fields::Field;
 use snarkvm_r1cs::errors::SynthesisError;
 
