@@ -14,6 +14,8 @@
 // You should have received a copy of the GNU General Public License
 // along with the snarkVM library. If not, see <https://www.gnu.org/licenses/>.
 
+use core::marker::PhantomData;
+
 use crate::{
     ahp::{
         indexer::CircuitInfo,
@@ -21,6 +23,7 @@ use crate::{
         AHPError,
         AHPForR1CS,
     },
+    marlin::MarlinMode,
     traits::FiatShamirRng,
 };
 use snarkvm_algorithms::fft::EvaluationDomain;
@@ -32,12 +35,12 @@ use snarkvm_polycommit::QuerySet;
 
 use rand_core::RngCore;
 
-impl<TargetField: PrimeField> AHPForR1CS<TargetField> {
+impl<TargetField: PrimeField, MM: MarlinMode> AHPForR1CS<TargetField, MM> {
     /// Output the first message and next round state.
     pub fn verifier_first_round<BaseField: PrimeField, R: FiatShamirRng<TargetField, BaseField>>(
         index_info: CircuitInfo<TargetField>,
         fs_rng: &mut R,
-    ) -> Result<(VerifierFirstMessage<TargetField>, VerifierState<TargetField>), AHPError> {
+    ) -> Result<(VerifierFirstMessage<TargetField>, VerifierState<TargetField, MM>), AHPError> {
         // Check that the R1CS is a square matrix.
         if index_info.num_constraints != index_info.num_variables {
             return Err(AHPError::NonSquareMatrix);
@@ -69,6 +72,7 @@ impl<TargetField: PrimeField> AHPForR1CS<TargetField> {
             first_round_message: Some(message),
             second_round_message: None,
             gamma: None,
+            mode: PhantomData,
         };
 
         Ok((message, new_state))
@@ -76,9 +80,9 @@ impl<TargetField: PrimeField> AHPForR1CS<TargetField> {
 
     /// Output the second message and next round state.
     pub fn verifier_second_round<BaseField: PrimeField, R: FiatShamirRng<TargetField, BaseField>>(
-        mut state: VerifierState<TargetField>,
+        mut state: VerifierState<TargetField, MM>,
         fs_rng: &mut R,
-    ) -> Result<(VerifierSecondMessage<TargetField>, VerifierState<TargetField>), AHPError> {
+    ) -> Result<(VerifierSecondMessage<TargetField>, VerifierState<TargetField, MM>), AHPError> {
         let elems = fs_rng.squeeze_nonnative_field_elements(1, OptimizationType::Weight)?;
         let beta = elems[0];
         assert!(!state.domain_h.evaluate_vanishing_polynomial(beta).is_zero());
@@ -91,9 +95,9 @@ impl<TargetField: PrimeField> AHPForR1CS<TargetField> {
 
     /// Output the third message and next round state.
     pub fn verifier_third_round<BaseField: PrimeField, R: FiatShamirRng<TargetField, BaseField>>(
-        mut state: VerifierState<TargetField>,
+        mut state: VerifierState<TargetField, MM>,
         fs_rng: &mut R,
-    ) -> Result<VerifierState<TargetField>, AHPError> {
+    ) -> Result<VerifierState<TargetField, MM>, AHPError> {
         let elems = fs_rng.squeeze_nonnative_field_elements(1, OptimizationType::Weight)?;
         let gamma = elems[0];
 
@@ -103,10 +107,10 @@ impl<TargetField: PrimeField> AHPForR1CS<TargetField> {
 
     /// Output the query state and next round state.
     pub fn verifier_query_set<'a, 'b, R: RngCore>(
-        state: VerifierState<TargetField>,
+        state: VerifierState<TargetField, MM>,
         _: &'a mut R,
-        with_vanishing: bool,
-    ) -> (QuerySet<'b, TargetField>, VerifierState<TargetField>) {
+    ) -> (QuerySet<'b, TargetField>, VerifierState<TargetField, MM>) {
+        let with_vanishing = MM::RECURSION;
         let alpha = state.first_round_message.unwrap().alpha;
         let beta = state.second_round_message.unwrap().beta;
         let gamma = state.gamma.unwrap();
@@ -139,10 +143,10 @@ impl<TargetField: PrimeField> AHPForR1CS<TargetField> {
         //  LinearCombination::new("z_b", vec![(F::one(), z_b)])
         //  LinearCombination::new("g_1", vec![(F::one(), g_1)], rhs::new(g_1_at_beta))
         //  LinearCombination::new("t", vec![(F::one(), t)])
-        query_set.insert(("g_1".into(), beta));
-        query_set.insert(("z_b".into(), beta));
-        query_set.insert(("t".into(), beta));
-        query_set.insert(("outer_sumcheck".into(), beta));
+        query_set.insert(("g_1".into(), ("beta".into(), beta)));
+        query_set.insert(("z_b".into(), ("beta".into(), beta)));
+        query_set.insert(("t".into(), ("beta".into(), beta)));
+        query_set.insert(("outer_sumcheck".into(), ("beta".into(), beta)));
 
         // For the second linear combination
         // Inner sumcheck test:
@@ -150,68 +154,47 @@ impl<TargetField: PrimeField> AHPForR1CS<TargetField> {
         // = a(gamma) - b(gamma) * (gamma g_2(gamma) + t(beta) / |K|)
         //
         // where
-        //   a(X) := sum_M (eta_M v_H(beta) v_H(alpha) val_M(X) prod_N (beta - row_N(X)) (alpha - col_N(X)))
-        //   b(X) := prod_M (beta - row_M(X)) (alpha - col_M(X))
-        //
-        // We define "n_denom" := prod_N (beta - row_N(X)) (alpha - col_N(X)))
+        //   a(X) := sum_M (eta_M v_H(beta) v_H(alpha) val_M(X))
+        //   b(X) := (beta - row(X)) (alpha - col(X))
         //
         // LinearCombination::new("g_2", vec![(F::one(), g_2)]);
         //
         // LinearCombination::new(
-        //     "a_denom".into(),
+        //     "denom".into(),
         //     vec![
         //         (alpha * beta, LCTerm::One),
-        //         (-alpha, "a_row"),
-        //         (-beta, "a_col"),
-        //         (F::one(), "a_row_col"),
-        // ]);
-        // LinearCombination::new(
-        //     "b_denom".into(),
-        //     vec![
-        //         (alpha * beta, LCTerm::One),
-        //         (-alpha, "b_row"),
-        //         (-beta, "b_col"),
-        //         (F::one(), "b_row_col"),
-        // ]);
-        // LinearCombination::new(
-        //     "c_denom".into(),
-        //     vec![
-        //         (alpha * beta, LCTerm::one()),
-        //         (-alpha, "c_row"),
-        //         (-beta, "c_col"),
-        //         (F::one(), "c_row_col"),
+        //         (-alpha, "row"),
+        //         (-beta, "col"),
+        //         (F::one(), "row_col"),
         // ]);
         //
         // LinearCombination::new(
         //     "a_poly".into(),
         //     vec![
-        //          (eta_a * b_denom_at_gamma * c_denom_at_gamma, "a_val".into()),
-        //          (eta_b * a_denom_at_gamma * c_denom_at_gamma, "b_val".into()),
-        //          (eta_c * b_denom_at_gamma * a_denom_at_gamma, "c_val".into()),
-        //     ],
-        // )
+        //          (eta_a * "a_val".into()),
+        //          (eta_b * "b_val".into()),
+        //          (eta_c * "c_val".into()),
+        //      ],
+        //  )
         //
         // let v_H_at_alpha = domain_h.evaluate_vanishing_polynomial(alpha);
         // let v_H_at_beta = domain_h.evaluate_vanishing_polynomial(beta);
         // let v_K_at_gamma = domain_k.evaluate_vanishing_polynomial(gamma);
         //
         // let a_poly_lc *= v_H_at_alpha * v_H_at_beta;
-        // let b_lc = LinearCombination::new("b_poly", vec![(a_denom_at_gamma * b_denom_at_gamma * c_denom_at_gamma, "one")]);
+        // let b_lc = denom
         // let h_lc = LinearCombination::new("b_poly", vec![(v_K_at_gamma, "h_2")]);
         //
         // // This LC is the only one that is evaluated:
         // let inner_sumcheck = a_poly_lc - (b_lc * (gamma * g_2_at_gamma + (t_at_beta / &k_size))) - h_lc
         // main_lc.set_label("inner_sumcheck");
-        query_set.insert(("g_2".into(), gamma));
-        query_set.insert(("a_denom".into(), gamma));
-        query_set.insert(("b_denom".into(), gamma));
-        query_set.insert(("c_denom".into(), gamma));
-        query_set.insert(("inner_sumcheck".into(), gamma));
+        query_set.insert(("g_2".into(), ("gamma".into(), gamma)));
+        query_set.insert(("inner_sumcheck".into(), ("gamma".into(), gamma)));
 
         if with_vanishing {
-            query_set.insert(("vanishing_poly_h_alpha".into(), alpha));
-            query_set.insert(("vanishing_poly_h_beta".into(), beta));
-            query_set.insert(("vanishing_poly_k_gamma".into(), gamma));
+            query_set.insert(("vanishing_poly_h_alpha".into(), ("alpha".into(), alpha)));
+            query_set.insert(("vanishing_poly_h_beta".into(), ("beta".into(), beta)));
+            query_set.insert(("vanishing_poly_k_gamma".into(), ("gamma".into(), gamma)));
         }
 
         (query_set, state)
