@@ -18,7 +18,7 @@ use snarkvm_fields::PrimeField;
 use snarkvm_gadgets::{nonnative::NonNativeFieldVar, traits::fields::FieldGadget};
 use snarkvm_r1cs::{ConstraintSystem, SynthesisError};
 
-use std::marker::PhantomData;
+use crate::PhantomData;
 
 /// A struct with methods to compute vanishing polynomial equations.
 pub struct AlgebraForAHP<F: PrimeField, CF: PrimeField> {
@@ -65,10 +65,10 @@ impl<TargetField: PrimeField, BaseField: PrimeField> AlgebraForAHP<TargetField,
         y_prepared: &NonNativeFieldVar<TargetField, BaseField>,
     ) -> Result<NonNativeFieldVar<TargetField, BaseField>, SynthesisError> {
         // Compute `x - y`.
-        let denominator = x.sub(cs.ns(|| "denominator"), &y)?;
+        let denominator = x.sub(cs.ns(|| "denominator"), y)?;
 
         // Compute `x_prepared - y_prepared`.
-        let numerator = x_prepared.sub(cs.ns(|| "numerator"), &y_prepared)?;
+        let numerator = x_prepared.sub(cs.ns(|| "numerator"), y_prepared)?;
         let denominator_invert = denominator.inverse(cs.ns(|| "denominator_inverse"))?;
 
         // Compute `(x - y) / (x_prepared - y_prepared)`.
