@@ -14,8 +14,7 @@
 // You should have received a copy of the GNU General Public License
 // along with the snarkVM library. If not, see <https://www.gnu.org/licenses/>.
 
-use std::marker::PhantomData;
-
+use snarkvm_algorithms::DefaultCapacityAlgebraicSponge;
 use snarkvm_fields::PrimeField;
 use snarkvm_gadgets::{
     bits::{Boolean, ToBitsBEGadget},
@@ -28,22 +27,21 @@ use snarkvm_gadgets::{
     },
     overhead,
     traits::{alloc::AllocGadget, fields::FieldGadget, integers::Integer},
+    DefaultCapacityAlgebraicSpongeVar,
 };
 use snarkvm_r1cs::{ConstraintSystem, ConstraintVariable, LinearCombination, SynthesisError};
 
-use crate::fiat_shamir::{
-    traits::{AlgebraicSpongeVar, FiatShamirRngVar},
-    AlgebraicSponge,
-    FiatShamirAlgebraicSpongeRng,
-};
+use crate::fiat_shamir::{traits::FiatShamirRngVar, FiatShamirAlgebraicSpongeRng};
+
+use crate::{PhantomData, Vec};
 
 /// Building the Fiat-Shamir sponge's gadget from any algebraic sponge's gadget.
 #[derive(Clone)]
 pub struct FiatShamirAlgebraicSpongeRngVar<
     TargetField: PrimeField,
     BaseField: PrimeField,
-    PS: AlgebraicSponge<BaseField>,
-    S: AlgebraicSpongeVar<BaseField, PS>,
+    PS: DefaultCapacityAlgebraicSponge<BaseField, 6>,
+    S: DefaultCapacityAlgebraicSpongeVar<BaseField, PS, 6>,
 > {
     /// Algebraic sponge gadget.
     pub s: S,
@@ -58,8 +56,8 @@ pub struct FiatShamirAlgebraicSpongeRngVar<
 impl<
     TargetField: PrimeField,
     BaseField: PrimeField,
-    PS: AlgebraicSponge<BaseField>,
-    S: AlgebraicSpongeVar<BaseField, PS>,
+    PS: DefaultCapacityAlgebraicSponge<BaseField, 6>,
+    S: DefaultCapacityAlgebraicSpongeVar<BaseField, PS, 6>,
 > FiatShamirAlgebraicSpongeRngVar<TargetField, BaseField, PS, S>
 {
     /// Compress every two elements if possible. Provides a vector of (limb, num_of_additions),
@@ -136,37 +134,37 @@ impl<
         for (i, elem) in src.iter().enumerate() {
             match elem {
                 NonNativeFieldVar::Constant(c) => {
-                    let v = AllocatedNonNativeFieldVar::<TargetField, BaseField>::alloc_constant(
-                        cs.ns(|| format!("alloc_constant_{}", i)),
-                        || Ok(c),
+                    let v = AllocatedNonNativeFieldVar::<TargetField, BaseField>::constant(
+                        &mut cs.ns(|| format!("alloc_constant_{}", i)),
+                        *c,
                     )?;
 
-                    for limb in v.limbs.iter() {
+                    for limb in v.limbs {
                         let num_of_additions_over_normal_form =
                             if v.num_of_additions_over_normal_form == BaseField::zero() {
                                 BaseField::one()
                             } else {
                                 v.num_of_additions_over_normal_form
                             };
-                        src_limbs.push((limb.clone(), num_of_additions_over_normal_form));
+                        src_limbs.push((limb, num_of_additions_over_normal_form));
                     }
                 }
                 NonNativeFieldVar::Var(v) => {
-                    for limb in v.limbs.iter() {
+                    for limb in v.limbs.clone() {
                         let num_of_additions_over_normal_form =
                             if v.num_of_additions_over_normal_form == BaseField::zero() {
                                 BaseField::one()
                             } else {
                                 v.num_of_additions_over_normal_form
                             };
-                        src_limbs.push((limb.clone(), num_of_additions_over_normal_form));
+                        src_limbs.push((limb, num_of_additions_over_normal_form));
                     }
                 }
             }
         }
 
         let dest_limbs = Self::compress_gadgets(cs.ns(|| "compress_gadgets"), &src_limbs, ty)?;
-        sponge.absorb(cs.ns(|| "absorb"), &dest_limbs)?;
+        sponge.absorb(cs.ns(|| "absorb"), dest_limbs.iter())?;
         Ok(())
     }
 
@@ -180,7 +178,7 @@ impl<
         let bits_per_element = BaseField::size_in_bits() - 1;
         let num_elements = (num_bits + bits_per_element - 1) / bits_per_element;
 
-        let src_elements = sponge.squeeze(cs.ns(|| "squeeze"), num_elements)?;
+        let src_elements = sponge.squeeze_field_elements(cs.ns(|| "squeeze"), num_elements)?;
         let mut dest_bits = Vec::<Boolean>::new();
 
         for (i, elem) in src_elements.iter().enumerate() {
@@ -311,14 +309,14 @@ impl<
 impl<
     TargetField: PrimeField,
     BaseField: PrimeField,
-    PS: AlgebraicSponge<BaseField>,
-    S: AlgebraicSpongeVar<BaseField, PS>,
+    PS: DefaultCapacityAlgebraicSponge<BaseField, 6>,
+    S: DefaultCapacityAlgebraicSpongeVar<BaseField, PS, 6>,
 > FiatShamirRngVar<TargetField, BaseField, FiatShamirAlgebraicSpongeRng<TargetField, BaseField, PS>>
     for FiatShamirAlgebraicSpongeRngVar<TargetField, BaseField, PS, S>
 {
     fn new<CS: ConstraintSystem<BaseField>>(cs: CS) -> Self {
         Self {
-            s: S::new(cs),
+            s: S::with_parameters(cs, &PS::sample_parameters()),
             _target_field: PhantomData,
             _base_field: PhantomData,
             _sponge: PhantomData,
@@ -343,7 +341,7 @@ impl<
         elems: &[NonNativeFieldVar<TargetField, BaseField>],
         ty: OptimizationType,
     ) -> Result<(), SynthesisError> {
-        Self::push_gadgets_to_sponge(cs, &mut self.s, &elems.to_vec(), ty)
+        Self::push_gadgets_to_sponge(cs, &mut self.s, elems, ty)
     }
 
     fn absorb_native_field_elements<CS: ConstraintSystem<BaseField>>(
@@ -351,7 +349,7 @@ impl<
         cs: CS,
         elems: &[FpGadget<BaseField>],
     ) -> Result<(), SynthesisError> {
-        self.s.absorb(cs, elems)?;
+        self.s.absorb(cs, elems.iter())?;
         Ok(())
     }
 
@@ -402,7 +400,7 @@ impl<
             cs.enforce(|| format!("enforce_constraint_{}", i), |lc| lc, |lc| lc, |_| lc);
         }
 
-        self.s.absorb(cs.ns(|| "absorb"), &gadgets)
+        self.s.absorb(cs.ns(|| "absorb"), gadgets.iter())
     }
 
     fn squeeze_native_field_elements<CS: ConstraintSystem<BaseField>>(
@@ -410,7 +408,7 @@ impl<
         cs: CS,
         num: usize,
     ) -> Result<Vec<FpGadget<BaseField>>, SynthesisError> {
-        self.s.squeeze(cs, num)
+        self.s.squeeze_field_elements(cs, num)
     }
 
     fn squeeze_field_elements<CS: ConstraintSystem<BaseField>>(
@@ -454,24 +452,21 @@ mod tests {
     use rand_chacha::ChaChaRng;
     use rand_core::SeedableRng;
 
-    use snarkvm_curves::bls12_377::Fr;
+    use snarkvm_curves::bls12_377::Fq;
     use snarkvm_fields::One;
     use snarkvm_gadgets::{bits::ToBitsLEGadget, traits::eq::EqGadget};
     use snarkvm_r1cs::TestConstraintSystem;
     use snarkvm_utilities::rand::UniformRand;
 
-    use crate::fiat_shamir::{
-        fiat_shamir_poseidon_sponge::PoseidonSponge,
-        fiat_shamir_poseidon_sponge_gadget::PoseidonSpongeVar,
-        traits::FiatShamirRng,
-    };
+    use crate::fiat_shamir::{traits::FiatShamirRng, PoseidonSpongeGadget};
+    use snarkvm_algorithms::crypto_hash::PoseidonSponge;
 
     use super::*;
 
-    type PS = PoseidonSponge<Fr>;
-    type PSGadget = PoseidonSpongeVar<Fr>;
-    type FS = FiatShamirAlgebraicSpongeRng<Fr, Fr, PS>;
-    type FSGadget = FiatShamirAlgebraicSpongeRngVar<Fr, Fr, PS, PSGadget>;
+    type PS = PoseidonSponge<Fq, 6, 1>;
+    type PSGadget = PoseidonSpongeGadget<Fq, 6, 1>;
+    type FS = FiatShamirAlgebraicSpongeRng<Fq, Fq, PS>;
+    type FSGadget = FiatShamirAlgebraicSpongeRngVar<Fq, Fq, PS, PSGadget>;
 
     const MAX_ELEMENTS: usize = 50;
     const MAX_ELEMENT_SIZE: usize = 100;
@@ -492,7 +487,7 @@ mod tests {
 
         let mut absorbed_rand_field_elems = Vec::new();
         for _ in 0..NUM_ABSORBED_RAND_FIELD_ELEMS {
-            absorbed_rand_field_elems.push(Fr::rand(rng));
+            absorbed_rand_field_elems.push(Fq::rand(rng));
         }
 
         let mut absorbed_rand_byte_elems = Vec::<Vec<u8>>::new();
@@ -517,7 +512,7 @@ mod tests {
             .unwrap();
 
         // fs_rng in the constraint world
-        let mut cs = TestConstraintSystem::<Fr>::new();
+        let mut cs = TestConstraintSystem::<Fq>::new();
         let mut fs_rng_gadget = FSGadget::new(cs.ns(|| "new"));
 
         let mut absorbed_rand_field_elems_gadgets = Vec::new();
@@ -599,72 +594,31 @@ mod tests {
         assert!(cs.is_satisfied());
     }
 
-    #[test]
-    fn test_fiat_shamir_algebraic_sponge_rng_constant() {
-        let mut rng = ChaChaRng::seed_from_u64(123456789u64);
-
-        for i in 0..ITERATIONS {
-            let mut cs = TestConstraintSystem::<Fr>::new();
-
-            // Generate random element.
-            let num_bytes: usize = rng.gen_range(0..MAX_ELEMENT_SIZE);
-            let element: Vec<u8> = (0..num_bytes).map(|_| u8::rand(&mut rng)).collect();
-
-            // Create a new FS rng.
-            let mut fs_rng = FS::new();
-            fs_rng.absorb_bytes(&element);
-
-            // Allocate an fs_rng gadget from the existing `fs_rng`.
-            let mut fs_rng_gadget = FSGadget::constant(cs.ns(|| format!("fs_rng_gadget_constant_{}", i)), &fs_rng);
-
-            // Get bits from the `fs_rng` and `fs_rng_gadget`.
-            let num_bits = num_bytes * 8;
-            let bits = FS::get_bits_from_sponge(&mut fs_rng.s, num_bits);
-            let bit_gadgets = FSGadget::get_booleans_from_sponge(
-                cs.ns(|| format!("get_booleans_from_sponge_{}", i)),
-                &mut fs_rng_gadget.s,
-                num_bits,
-            )
-            .unwrap();
-
-            // Check that the bit results are equivalent.
-            for (j, (bit_gadget, bit)) in bit_gadgets.iter().zip(bits).enumerate() {
-                // Allocate a boolean from the native bit.
-                let alloc_boolean = Boolean::alloc(cs.ns(|| format!("alloc_boolean_{}_{}", i, j)), || Ok(bit)).unwrap();
-
-                // Check that the boolean gadgets are equivalent.
-                bit_gadget
-                    .enforce_equal(cs.ns(|| format!("enforce_equal_bit_{}_{}", i, j)), &alloc_boolean)
-                    .unwrap();
-            }
-        }
-    }
-
     #[test]
     fn test_compress_gadgets_weight_optimized() {
         let mut rng = ChaChaRng::seed_from_u64(123456789u64);
 
         for i in 0..ITERATIONS {
-            let mut cs = TestConstraintSystem::<Fr>::new();
+            let mut cs = TestConstraintSystem::<Fq>::new();
 
             // Generate random elements.
             let num_elements: usize = rng.gen_range(0..MAX_ELEMENT_SIZE);
-            let elements: Vec<_> = (0..num_elements).map(|_| Fr::rand(&mut rng)).collect();
+            let elements: Vec<_> = (0..num_elements).map(|_| Fq::rand(&mut rng)).collect();
 
             // Construct elements limb representations
-            let mut element_limbs = Vec::<(Fr, Fr)>::new();
-            let mut element_limb_gadgets = Vec::<(FpGadget<Fr>, Fr)>::new();
+            let mut element_limbs = Vec::<(Fq, Fq)>::new();
+            let mut element_limb_gadgets = Vec::<(FpGadget<Fq>, Fq)>::new();
 
             for (j, elem) in elements.iter().enumerate() {
                 let limbs =
-                    AllocatedNonNativeFieldVar::<Fr, Fr>::get_limbs_representations(elem, OptimizationType::Weight)
+                    AllocatedNonNativeFieldVar::<Fq, Fq>::get_limbs_representations(elem, OptimizationType::Weight)
                         .unwrap();
                 for (k, limb) in limbs.iter().enumerate() {
                     let allocated_limb =
                         FpGadget::alloc(cs.ns(|| format!("alloc_limb_{}_{}_{}", i, j, k)), || Ok(limb)).unwrap();
 
-                    element_limbs.push((*limb, Fr::one()));
-                    element_limb_gadgets.push((allocated_limb, Fr::one()));
+                    element_limbs.push((*limb, Fq::one()));
+                    element_limb_gadgets.push((allocated_limb, Fq::one()));
                     // Specifically set to one, since most gadgets in the constraint world would not have zero noise (due to the relatively weak normal form testing in `alloc`)
                 }
             }
@@ -697,18 +651,18 @@ mod tests {
         let mut rng = ChaChaRng::seed_from_u64(123456789u64);
 
         for i in 0..ITERATIONS {
-            let mut cs = TestConstraintSystem::<Fr>::new();
+            let mut cs = TestConstraintSystem::<Fq>::new();
 
             // Generate random elements.
             let num_elements: usize = rng.gen_range(0..MAX_ELEMENT_SIZE);
-            let elements: Vec<_> = (0..num_elements).map(|_| Fr::rand(&mut rng)).collect();
+            let elements: Vec<_> = (0..num_elements).map(|_| Fq::rand(&mut rng)).collect();
 
             // Construct elements limb representations
-            let mut element_limbs = Vec::<(Fr, Fr)>::new();
-            let mut element_limb_gadgets = Vec::<(FpGadget<Fr>, Fr)>::new();
+            let mut element_limbs = Vec::<(Fq, Fq)>::new();
+            let mut element_limb_gadgets = Vec::<(FpGadget<Fq>, Fq)>::new();
 
             for (j, elem) in elements.iter().enumerate() {
-                let limbs = AllocatedNonNativeFieldVar::<Fr, Fr>::get_limbs_representations(
+                let limbs = AllocatedNonNativeFieldVar::<Fq, Fq>::get_limbs_representations(
                     elem,
                     OptimizationType::Constraints,
                 )
@@ -717,8 +671,8 @@ mod tests {
                     let allocated_limb =
                         FpGadget::alloc(cs.ns(|| format!("alloc_limb_{}_{}_{}", i, j, k)), || Ok(limb)).unwrap();
 
-                    element_limbs.push((*limb, Fr::one()));
-                    element_limb_gadgets.push((allocated_limb, Fr::one()));
+                    element_limbs.push((*limb, Fq::one()));
+                    element_limb_gadgets.push((allocated_limb, Fq::one()));
                     // Specifically set to one, since most gadgets in the constraint world would not have zero noise (due to the relatively weak normal form testing in `alloc`)
                 }
             }
@@ -751,7 +705,7 @@ mod tests {
         let mut rng = ChaChaRng::seed_from_u64(123456789u64);
 
         for i in 0..ITERATIONS {
-            let mut cs = TestConstraintSystem::<Fr>::new();
+            let mut cs = TestConstraintSystem::<Fq>::new();
 
             // Create a new FS rng and FS rng gadget.
             let mut fs_rng = FS::new();
@@ -804,7 +758,7 @@ mod tests {
         let mut rng = ChaChaRng::seed_from_u64(123456789u64);
 
         for i in 0..ITERATIONS {
-            let mut cs = TestConstraintSystem::<Fr>::new();
+            let mut cs = TestConstraintSystem::<Fq>::new();
 
             // Create a new FS rng.
             let mut fs_rng = FS::new();
@@ -814,7 +768,7 @@ mod tests {
 
             // Generate random elements.
             let num_elements: usize = rng.gen_range(0..MAX_ELEMENTS);
-            let elements: Vec<_> = (0..num_elements).map(|_| Fr::rand(&mut rng)).collect();
+            let elements: Vec<_> = (0..num_elements).map(|_| Fq::rand(&mut rng)).collect();
 
             let mut element_gadgets = vec![];
             for (j, element) in elements.iter().enumerate() {
@@ -858,7 +812,7 @@ mod tests {
         let mut rng = ChaChaRng::seed_from_u64(123456789u64);
 
         for i in 0..ITERATIONS {
-            let mut cs = TestConstraintSystem::<Fr>::new();
+            let mut cs = TestConstraintSystem::<Fq>::new();
 
             // Create a new FS rng.
             let mut fs_rng = FS::new();
@@ -868,7 +822,7 @@ mod tests {
 
             // Generate random elements.
             let num_elements: usize = rng.gen_range(0..MAX_ELEMENTS);
-            let elements: Vec<_> = (0..num_elements).map(|_| Fr::rand(&mut rng)).collect();
+            let elements: Vec<_> = (0..num_elements).map(|_| Fq::rand(&mut rng)).collect();
 
             let mut element_gadgets = vec![];
             for (j, element) in elements.iter().enumerate() {
@@ -920,7 +874,7 @@ mod tests {
         let mut rng = ChaChaRng::seed_from_u64(123456789u64);
 
         for i in 0..ITERATIONS {
-            let mut cs = TestConstraintSystem::<Fr>::new();
+            let mut cs = TestConstraintSystem::<Fq>::new();
 
             // Create a new FS rng.
             let mut fs_rng = FS::new();
@@ -930,7 +884,7 @@ mod tests {
 
             // Generate random elements.
             let num_elements: usize = rng.gen_range(0..MAX_ELEMENTS);
-            let elements: Vec<_> = (0..num_elements).map(|_| Fr::rand(&mut rng)).collect();
+            let elements: Vec<_> = (0..num_elements).map(|_| Fq::rand(&mut rng)).collect();
 
             let mut element_gadgets = vec![];
             for (j, element) in elements.iter().enumerate() {
@@ -998,7 +952,7 @@ mod tests {
         let mut rng = ChaChaRng::seed_from_u64(123456789u64);
 
         for i in 0..ITERATIONS {
-            let mut cs = TestConstraintSystem::<Fr>::new();
+            let mut cs = TestConstraintSystem::<Fq>::new();
 
             // Create a new FS rng.
             let mut fs_rng = FS::new();
@@ -1008,7 +962,7 @@ mod tests {
 
             // Generate random elements.
             let num_elements: usize = rng.gen_range(0..MAX_ELEMENTS);
-            let elements: Vec<_> = (0..num_elements).map(|_| Fr::rand(&mut rng)).collect();
+            let elements: Vec<_> = (0..num_elements).map(|_| Fq::rand(&mut rng)).collect();
 
             let mut element_gadgets = vec![];
             for (j, element) in elements.iter().enumerate() {
@@ -1061,7 +1015,7 @@ mod tests {
         let mut rng = ChaChaRng::seed_from_u64(123456789u64);
 
         for i in 0..ITERATIONS {
-            let mut cs = TestConstraintSystem::<Fr>::new();
+            let mut cs = TestConstraintSystem::<Fq>::new();
 
             // Create a new FS rng.
             let mut fs_rng = FS::new();
@@ -1071,7 +1025,7 @@ mod tests {
 
             // Generate random elements.
             let num_elements: usize = rng.gen_range(0..MAX_ELEMENTS);
-            let elements: Vec<_> = (0..num_elements).map(|_| Fr::rand(&mut rng)).collect();
+            let elements: Vec<_> = (0..num_elements).map(|_| Fq::rand(&mut rng)).collect();
 
             let mut element_gadgets = vec![];
             for (j, element) in elements.iter().enumerate() {
