@@ -14,7 +14,7 @@
 // You should have received a copy of the GNU General Public License
 // along with the snarkVM library. If not, see <https://www.gnu.org/licenses/>.
 
-use crate::traits::FiatShamirRng;
+use crate::{traits::FiatShamirRng, Vec};
 use snarkvm_fields::PrimeField;
 use snarkvm_gadgets::{
     bits::Boolean,
