@@ -23,14 +23,9 @@
 //! is the same as the number of constraints (i.e., where the constraint
 //! matrices are square). Furthermore, Marlin only supports instances where the
 //! public inputs are of size one less than a power of 2 (i.e., 2^n - 1).
-#![allow(clippy::module_inception)]
-#![deny(unused_import_braces, trivial_casts, bare_trait_objects)]
-#![deny(unused_qualifications, variant_size_differences, stable_features)]
-#![deny(non_shorthand_field_patterns, unused_attributes)]
-#![deny(renamed_and_removed_lints, unused_allocation, unused_comparisons)]
-#![deny(const_err, unused_must_use, unused_mut, private_in_public)]
-#![deny(unused_extern_crates, trivial_numeric_casts)]
 #![forbid(unsafe_code)]
+#![allow(clippy::module_inception)]
+#![allow(clippy::type_complexity)]
 
 #[macro_use]
 extern crate snarkvm_profiler;
@@ -42,20 +37,32 @@ extern crate alloc;
 #[rustfmt::skip]
 #[cfg(not(feature = "std"))]
 use alloc::{
-    collections::BTreeMap,
-    marker::PhantomData,
+    collections::{BTreeSet, BTreeMap},
     string::{String, ToString},
     vec::Vec,
 };
 
+#[cfg(not(feature = "std"))]
+use core::marker::PhantomData;
+
+#[cfg(not(feature = "std"))]
+use snarkvm_utilities::io::{
+    Read,
+    Result as IoResult,
+    Write,
+    {self},
+};
+
 #[rustfmt::skip]
 #[cfg(feature = "std")]
 use std::{
-    collections::BTreeMap,
+    collections::{BTreeSet, BTreeMap},
     marker::PhantomData,
     string::{String, ToString},
     vec::Vec,
+    io::{Read, Write, Result as IoResult, {self}},
 };
+use snarkvm_polycommit::PolynomialCommitment;
 
 #[cfg(not(feature = "std"))]
 macro_rules! eprintln {
@@ -63,6 +70,9 @@ macro_rules! eprintln {
     ($($arg: tt)*) => {};
 }
 
+#[macro_use]
+extern crate snarkvm_utilities;
+
 /// Implements an Algebraic Holographic Proof (AHP) for the R1CS indexed relation.
 pub mod ahp;
 pub use ahp::*;
@@ -73,14 +83,6 @@ pub mod constraints;
 /// Implements the base Marlin zkSNARK proof system.
 pub mod marlin;
 
-/// The Marlin public parameters for a given circuit.
-pub mod parameters;
-pub use parameters::*;
-
-/// Implements the snarkVM-compatible Marlin SNARK interface.
-pub mod snark;
-pub use snark::*;
-
 /// RNGs for the Marlin SNARK.
 pub mod fiat_shamir;
 pub use fiat_shamir::*;
