@@ -17,25 +17,43 @@
 use crate::{
     ahp::{AHPError, AHPForR1CS, EvaluationsProvider},
     fiat_shamir::traits::FiatShamirRng,
-    marlin::{compute_vk_hash, CircuitProvingKey, CircuitVerifyingKey, MarlinError, MarlinMode, Proof, UniversalSRS},
+    marlin::{CircuitProvingKey, CircuitVerifyingKey, MarlinError, MarlinMode, Proof, UniversalSRS},
+    prover::ProverConstraintSystem,
+    String,
+    ToString,
+    Vec,
 };
 use snarkvm_algorithms::fft::EvaluationDomain;
-use snarkvm_fields::{PrimeField, ToConstraintField};
+use snarkvm_fields::PrimeField;
 use snarkvm_gadgets::nonnative::params::OptimizationType;
-use snarkvm_polycommit::{Evaluations, LabeledCommitment, LabeledPolynomial, PCUniversalParams, PolynomialCommitment};
+use snarkvm_polycommit::{
+    Evaluations,
+    LabeledCommitment,
+    LabeledPolynomial,
+    PCProof,
+    PCRandomness,
+    PCUniversalParams,
+    PolynomialCommitment,
+};
 use snarkvm_r1cs::{ConstraintSynthesizer, SynthesisError};
 use snarkvm_utilities::{to_bytes_le, ToBytes};
 
+#[cfg(not(feature = "std"))]
+use snarkvm_utilities::println;
+
 use crate::marlin::PreparedCircuitVerifyingKey;
-use core::marker::PhantomData;
+use core::{
+    marker::PhantomData,
+    sync::atomic::{AtomicBool, Ordering},
+};
 use rand_core::RngCore;
-use std::sync::atomic::{AtomicBool, Ordering};
 
 /// The Marlin proof system.
+#[derive(Clone, Debug)]
 pub struct MarlinSNARK<
     TargetField: PrimeField,
     BaseField: PrimeField,
-    PC: PolynomialCommitment<TargetField>,
+    PC: PolynomialCommitment<TargetField, BaseField>,
     FS: FiatShamirRng<TargetField, BaseField>,
     MM: MarlinMode,
 >(
@@ -49,13 +67,10 @@ pub struct MarlinSNARK<
 impl<
     TargetField: PrimeField,
     BaseField: PrimeField,
-    PC: PolynomialCommitment<TargetField>,
+    PC: PolynomialCommitment<TargetField, BaseField>,
     FS: FiatShamirRng<TargetField, BaseField>,
     MM: MarlinMode,
 > MarlinSNARK<TargetField, BaseField, PC, FS, MM>
-where
-    PC::VerifierKey: ToConstraintField<BaseField>,
-    PC::Commitment: ToConstraintField<BaseField>,
 {
     /// The personalization string for this protocol.
     /// Used to personalize the Fiat-Shamir RNG.
@@ -63,18 +78,10 @@ where
 
     /// Generates the universal proving and verifying keys for the argument system.
     pub fn universal_setup<R: RngCore>(
-        num_constraints: usize,
-        num_variables: usize,
-        num_non_zero: usize,
+        max_degree: usize,
         rng: &mut R,
-    ) -> Result<UniversalSRS<TargetField, PC>, MarlinError> {
-        let max_degree = AHPForR1CS::<TargetField>::max_degree(num_constraints, num_variables, num_non_zero)?;
-        let setup_time = start_timer!(|| {
-            format!(
-                "Marlin::UniversalSetup with max_degree {}, computed for a maximum of {} constraints, {} vars, {} non_zero",
-                max_degree, num_constraints, num_variables, num_non_zero,
-            )
-        });
+    ) -> Result<UniversalSRS<TargetField, BaseField, PC>, MarlinError> {
+        let setup_time = start_timer!(|| { format!("Marlin::UniversalSetup with max_degree {}", max_degree,) });
 
         let srs = PC::setup(max_degree, rng).map_err(Into::into);
         end_timer!(setup_time);
@@ -87,16 +94,20 @@ where
     pub fn circuit_specific_setup<C: ConstraintSynthesizer<TargetField>, R: RngCore>(
         c: &C,
         rng: &mut R,
-    ) -> Result<(CircuitProvingKey<TargetField, PC>, CircuitVerifyingKey<TargetField, PC>), MarlinError> {
+    ) -> Result<
+        (
+            CircuitProvingKey<TargetField, BaseField, PC, MM>,
+            CircuitVerifyingKey<TargetField, BaseField, PC, MM>,
+        ),
+        MarlinError,
+    > {
         let index_time = start_timer!(|| "Marlin::CircuitSpecificSetup");
 
-        let for_recursion = MM::RECURSION;
-
         // TODO: Add check that c is in the correct mode.
-        let circuit = AHPForR1CS::index(c)?;
+        let circuit = AHPForR1CS::<_, MM>::index(c)?;
         let srs = PC::setup(circuit.max_degree(), rng)?;
 
-        let coeff_support = AHPForR1CS::get_degree_bounds(&circuit.index_info);
+        let coeff_support = AHPForR1CS::<_, MM>::get_degree_bounds(&circuit.index_info);
 
         // Marlin only needs degree 2 random polynomials
         let supported_hiding_bound = 1;
@@ -104,7 +115,7 @@ where
             PC::trim(&srs, circuit.max_degree(), supported_hiding_bound, Some(&coeff_support))?;
 
         let mut vanishing_polys = vec![];
-        if for_recursion {
+        if MM::RECURSION {
             let domain_h = EvaluationDomain::new(circuit.index_info.num_constraints)
                 .ok_or(SynthesisError::PolynomialDegreeTooLarge)?;
             let domain_k = EvaluationDomain::new(circuit.index_info.num_non_zero)
@@ -139,6 +150,7 @@ where
             circuit_info: circuit.index_info,
             circuit_commitments,
             verifier_key,
+            mode: PhantomData,
         };
 
         let index_pk = CircuitProvingKey {
@@ -157,15 +169,19 @@ where
     /// This is a deterministic algorithm that anyone can rerun.
     #[allow(clippy::type_complexity)]
     pub fn circuit_setup<C: ConstraintSynthesizer<TargetField>>(
-        universal_srs: &UniversalSRS<TargetField, PC>,
+        universal_srs: &UniversalSRS<TargetField, BaseField, PC>,
         circuit: &C,
-    ) -> Result<(CircuitProvingKey<TargetField, PC>, CircuitVerifyingKey<TargetField, PC>), MarlinError> {
+    ) -> Result<
+        (
+            CircuitProvingKey<TargetField, BaseField, PC, MM>,
+            CircuitVerifyingKey<TargetField, BaseField, PC, MM>,
+        ),
+        MarlinError,
+    > {
         let index_time = start_timer!(|| "Marlin::CircuitSetup");
 
-        let is_recursion = MM::RECURSION;
-
         // TODO: Add check that c is in the correct mode.
-        let index = AHPForR1CS::index(circuit)?;
+        let index = AHPForR1CS::<_, MM>::index(circuit)?;
         if universal_srs.max_degree() < index.max_degree() {
             return Err(MarlinError::IndexTooLarge(
                 universal_srs.max_degree(),
@@ -173,19 +189,19 @@ where
             ));
         }
 
-        let coefficient_support = AHPForR1CS::get_degree_bounds(&index.index_info);
+        let coefficient_support = AHPForR1CS::<_, MM>::get_degree_bounds(&index.index_info);
 
         // Marlin only needs degree 2 random polynomials.
         let supported_hiding_bound = 1;
         let (committer_key, verifier_key) = PC::trim(
-            &universal_srs,
+            universal_srs,
             index.max_degree(),
             supported_hiding_bound,
             Some(&coefficient_support),
         )?;
 
         let mut vanishing_polynomials = vec![];
-        if is_recursion {
+        if MM::RECURSION {
             let domain_h = EvaluationDomain::new(index.index_info.num_constraints)
                 .ok_or(SynthesisError::PolynomialDegreeTooLarge)?;
             let domain_k =
@@ -220,6 +236,7 @@ where
             circuit_info: index.index_info,
             circuit_commitments,
             verifier_key,
+            mode: PhantomData,
         };
 
         let circuit_proving_key = CircuitProvingKey {
@@ -236,48 +253,43 @@ where
 
     /// Create a zkSNARK asserting that the constraint system is satisfied.
     pub fn prove<C: ConstraintSynthesizer<TargetField>, R: RngCore>(
-        circuit_proving_key: &CircuitProvingKey<TargetField, PC>,
+        circuit_proving_key: &CircuitProvingKey<TargetField, BaseField, PC, MM>,
         circuit: &C,
         zk_rng: &mut R,
-    ) -> Result<Proof<TargetField, PC>, MarlinError> {
+    ) -> Result<Proof<TargetField, BaseField, PC>, MarlinError> {
         Self::prove_with_terminator(circuit_proving_key, circuit, &AtomicBool::new(false), zk_rng)
     }
 
     /// Same as [`prove`] with an added termination flag, [`terminator`].
     pub fn prove_with_terminator<C: ConstraintSynthesizer<TargetField>, R: RngCore>(
-        circuit_proving_key: &CircuitProvingKey<TargetField, PC>,
+        circuit_proving_key: &CircuitProvingKey<TargetField, BaseField, PC, MM>,
         circuit: &C,
         terminator: &AtomicBool,
         zk_rng: &mut R,
-    ) -> Result<Proof<TargetField, PC>, MarlinError> {
+    ) -> Result<Proof<TargetField, BaseField, PC>, MarlinError> {
         let prover_time = start_timer!(|| "Marlin::Prover");
         // TODO: Add check that c is in the correct mode.
 
-        let is_recursion = MM::RECURSION;
-
         if terminator.load(Ordering::Relaxed) {
             return Err(MarlinError::Terminated);
         }
 
-        let prover_init_state = AHPForR1CS::prover_init(&circuit_proving_key.circuit, circuit)?;
+        let prover_init_state = AHPForR1CS::<_, MM>::prover_init(&circuit_proving_key.circuit, circuit)?;
         let public_input = prover_init_state.public_input();
+        let padded_public_input = prover_init_state.padded_public_input();
 
         let mut fs_rng = FS::new();
 
-        let hiding = !is_recursion;
-
-        if is_recursion {
+        if MM::RECURSION {
             fs_rng.absorb_bytes(&to_bytes_le![&Self::PROTOCOL_NAME].unwrap());
-            fs_rng.absorb_native_field_elements(&compute_vk_hash::<TargetField, BaseField, PC, FS>(
-                &circuit_proving_key.circuit_verifying_key,
-            )?);
-            fs_rng.absorb_nonnative_field_elements(&public_input, OptimizationType::Weight);
+            fs_rng.absorb_native_field_elements(&circuit_proving_key.circuit_verifying_key.circuit_commitments);
+            fs_rng.absorb_nonnative_field_elements(padded_public_input, OptimizationType::Weight);
         } else {
             fs_rng.absorb_bytes(
                 &to_bytes_le![
                     &Self::PROTOCOL_NAME,
                     &circuit_proving_key.circuit_verifying_key,
-                    &public_input
+                    padded_public_input
                 ]
                 .unwrap(),
             );
@@ -291,7 +303,7 @@ where
         }
 
         let (prover_first_message, prover_first_oracles, prover_state) =
-            AHPForR1CS::prover_first_round(prover_init_state, zk_rng, hiding)?;
+            AHPForR1CS::<_, MM>::prover_first_round(prover_init_state, zk_rng)?;
 
         if terminator.load(Ordering::Relaxed) {
             return Err(MarlinError::Terminated);
@@ -305,7 +317,7 @@ where
         )?;
         end_timer!(first_round_comm_time);
 
-        if is_recursion {
+        if MM::RECURSION {
             fs_rng.absorb_native_field_elements(&first_commitments);
             if !prover_first_message.field_elements.is_empty() {
                 fs_rng.absorb_nonnative_field_elements(&prover_first_message.field_elements, OptimizationType::Weight);
@@ -318,8 +330,10 @@ where
             return Err(MarlinError::Terminated);
         }
 
-        let (verifier_first_message, verifier_state) =
-            AHPForR1CS::verifier_first_round(circuit_proving_key.circuit_verifying_key.circuit_info, &mut fs_rng)?;
+        let (verifier_first_message, verifier_state) = AHPForR1CS::<_, MM>::verifier_first_round(
+            circuit_proving_key.circuit_verifying_key.circuit_info,
+            &mut fs_rng,
+        )?;
         // --------------------------------------------------------------------
 
         // --------------------------------------------------------------------
@@ -330,7 +344,7 @@ where
         }
 
         let (prover_second_message, prover_second_oracles, prover_state) =
-            AHPForR1CS::prover_second_round(&verifier_first_message, prover_state, zk_rng, hiding);
+            AHPForR1CS::<_, MM>::prover_second_round(&verifier_first_message, prover_state, zk_rng);
 
         let second_round_comm_time = start_timer!(|| "Committing to second round polys");
         let (second_commitments, second_commitment_randomnesses) = PC::commit_with_terminator(
@@ -341,7 +355,7 @@ where
         )?;
         end_timer!(second_round_comm_time);
 
-        if is_recursion {
+        if MM::RECURSION {
             fs_rng.absorb_native_field_elements(&second_commitments);
             if !prover_second_message.field_elements.is_empty() {
                 fs_rng.absorb_nonnative_field_elements(&prover_second_message.field_elements, OptimizationType::Weight);
@@ -354,7 +368,8 @@ where
             return Err(MarlinError::Terminated);
         }
 
-        let (verifier_second_msg, verifier_state) = AHPForR1CS::verifier_second_round(verifier_state, &mut fs_rng)?;
+        let (verifier_second_msg, verifier_state) =
+            AHPForR1CS::<_, MM>::verifier_second_round(verifier_state, &mut fs_rng)?;
         // --------------------------------------------------------------------
 
         // --------------------------------------------------------------------
@@ -365,7 +380,7 @@ where
         }
 
         let (prover_third_message, prover_third_oracles) =
-            AHPForR1CS::prover_third_round(&verifier_second_msg, prover_state, zk_rng)?;
+            AHPForR1CS::<_, MM>::prover_third_round(&verifier_second_msg, prover_state, zk_rng)?;
 
         let third_round_comm_time = start_timer!(|| "Committing to third round polys");
         let (third_commitments, third_commitment_randomnesses) = PC::commit_with_terminator(
@@ -376,7 +391,7 @@ where
         )?;
         end_timer!(third_round_comm_time);
 
-        if is_recursion {
+        if MM::RECURSION {
             fs_rng.absorb_native_field_elements(&third_commitments);
             if !prover_third_message.field_elements.is_empty() {
                 fs_rng.absorb_nonnative_field_elements(&prover_third_message.field_elements, OptimizationType::Weight);
@@ -385,18 +400,14 @@ where
             fs_rng.absorb_bytes(&to_bytes_le![third_commitments, prover_third_message].unwrap());
         }
 
-        if terminator.load(Ordering::Relaxed) {
-            return Err(MarlinError::Terminated);
-        }
-
-        let verifier_state = AHPForR1CS::verifier_third_round(verifier_state, &mut fs_rng)?;
+        let verifier_state = AHPForR1CS::<_, MM>::verifier_third_round(verifier_state, &mut fs_rng)?;
         // --------------------------------------------------------------------
 
         if terminator.load(Ordering::Relaxed) {
             return Err(MarlinError::Terminated);
         }
 
-        let vanishing_polys = if is_recursion {
+        let vanishing_polys = if MM::RECURSION {
             let domain_h = EvaluationDomain::new(circuit_proving_key.circuit.index_info.num_constraints)
                 .ok_or(SynthesisError::PolynomialDegreeTooLarge)?;
             let domain_k = EvaluationDomain::new(circuit_proving_key.circuit.index_info.num_non_zero)
@@ -425,36 +436,30 @@ where
             .circuit
             .iter() // 12 items
             .chain(vanishing_polys.iter()) // 0 or 2 items
-            .chain(prover_first_oracles.iter()) // 4 items
+            .chain(prover_first_oracles.iter()) // 3 or 4 items
             .chain(prover_second_oracles.iter())// 3 items
             .chain(prover_third_oracles.iter())// 2 items
             .collect();
 
-        // Sanity check, whose length should be updated if the underlying structs are updated.
-        match is_recursion {
-            true => assert_eq!(23, polynomials.len()),
-            false => assert_eq!(21, polynomials.len()),
-        };
-
         if terminator.load(Ordering::Relaxed) {
             return Err(MarlinError::Terminated);
         }
 
+        // Sanity check, whose length should be updated if the underlying structs are updated.
+        assert_eq!(
+            polynomials.len(),
+            AHPForR1CS::<TargetField, MM>::polynomial_labels().count()
+        );
+
         // Gather commitments in one vector.
         #[rustfmt::skip]
-            let commitments = vec![
+        let commitments = vec![
             first_commitments.iter().map(|p| p.commitment()).cloned().collect(),
             second_commitments.iter().map(|p| p.commitment()).cloned().collect(),
             third_commitments.iter().map(|p| p.commitment()).cloned().collect(),
         ];
 
-        let indexer_polynomials = if is_recursion {
-            AHPForR1CS::<TargetField>::INDEXER_POLYNOMIALS_WITH_VANISHING
-                .clone()
-                .to_vec()
-        } else {
-            AHPForR1CS::<TargetField>::INDEXER_POLYNOMIALS.clone().to_vec()
-        };
+        let indexer_polynomials = AHPForR1CS::<TargetField, MM>::indexer_polynomials();
 
         let labeled_commitments: Vec<_> = circuit_proving_key
             .circuit_verifying_key
@@ -477,14 +482,14 @@ where
             .chain(third_commitment_randomnesses)
             .collect();
 
-        if terminator.load(Ordering::Relaxed) {
-            return Err(MarlinError::Terminated);
+        if !MM::ZK {
+            let empty_randomness = PC::Randomness::empty();
+            assert!(commitment_randomnesses.iter().all(|r| r == &empty_randomness));
         }
 
         // Compute the AHP verifier's query set.
-        let (query_set, verifier_state) = AHPForR1CS::verifier_query_set(verifier_state, &mut fs_rng, is_recursion);
-        let lc_s =
-            AHPForR1CS::construct_linear_combinations(&public_input, &polynomials, &verifier_state, is_recursion)?;
+        let (query_set, verifier_state) = AHPForR1CS::<_, MM>::verifier_query_set(verifier_state, &mut fs_rng);
+        let lc_s = AHPForR1CS::<_, MM>::construct_linear_combinations(&public_input, &polynomials, &verifier_state)?;
 
         if terminator.load(Ordering::Relaxed) {
             return Err(MarlinError::Terminated);
@@ -492,21 +497,17 @@ where
 
         let eval_time = start_timer!(|| "Evaluating linear combinations over query set");
         let mut evaluations_unsorted = Vec::new();
-        for (label, point) in &query_set {
+        for (label, (_point_name, point)) in &query_set {
             let lc = lc_s
                 .iter()
                 .find(|lc| &lc.label == label)
                 .ok_or_else(|| AHPError::MissingEval(label.to_string()))?;
-            let evaluation = polynomials.get_lc_eval(&lc, *point)?;
-            if !AHPForR1CS::<TargetField>::LC_WITH_ZERO_EVAL.contains(&lc.label.as_ref()) {
+            let evaluation = polynomials.get_lc_eval(lc, *point)?;
+            if !AHPForR1CS::<TargetField, MM>::LC_WITH_ZERO_EVAL.contains(&lc.label.as_ref()) {
                 evaluations_unsorted.push((label.to_string(), evaluation));
             }
         }
 
-        if terminator.load(Ordering::Relaxed) {
-            return Err(MarlinError::Terminated);
-        }
-
         evaluations_unsorted.sort_by(|a, b| a.0.cmp(&b.0));
         let evaluations = evaluations_unsorted.iter().map(|x| x.1).collect::<Vec<TargetField>>();
         end_timer!(eval_time);
@@ -515,14 +516,14 @@ where
             return Err(MarlinError::Terminated);
         }
 
-        if is_recursion {
+        if MM::RECURSION {
             fs_rng.absorb_nonnative_field_elements(&evaluations, OptimizationType::Weight);
         } else {
             fs_rng.absorb_bytes(&to_bytes_le![&evaluations].unwrap());
         }
 
-        let pc_proof = if is_recursion {
-            let num_open_challenges: usize = 7;
+        let pc_proof = if MM::RECURSION {
+            let num_open_challenges: usize = 5;
 
             let mut opening_challenges = Vec::new();
             opening_challenges.append(&mut fs_rng.squeeze_128_bits_nonnative_field_elements(num_open_challenges)?);
@@ -561,6 +562,7 @@ where
         let prover_messages = vec![prover_first_message, prover_second_message, prover_third_message];
 
         let proof = Proof::new(commitments, evaluations, prover_messages, pc_proof);
+        assert_eq!(proof.pc_proof.is_hiding(), MM::ZK);
         proof.print_size_info();
         end_timer!(prover_time);
 
@@ -570,13 +572,40 @@ where
     /// Verify that a proof for the constraint system defined by `C` asserts that
     /// all constraints are satisfied.
     pub fn verify(
-        circuit_verifying_key: &CircuitVerifyingKey<TargetField, PC>,
+        circuit_verifying_key: &CircuitVerifyingKey<TargetField, BaseField, PC, MM>,
         public_input: &[TargetField],
-        proof: &Proof<TargetField, PC>,
+        proof: &Proof<TargetField, BaseField, PC>,
+    ) -> Result<bool, MarlinError> {
+        Self::verify_with_fs_parameters(circuit_verifying_key, &FS::sample_params(), public_input, proof)
+    }
+
+    /// Verify that a proof for the constraint system defined by `C` asserts that
+    /// all constraints are satisfied.
+    pub fn verify_with_fs_parameters(
+        circuit_verifying_key: &CircuitVerifyingKey<TargetField, BaseField, PC, MM>,
+        fs_parameters: &FS::Parameters,
+        public_input: &[TargetField],
+        proof: &Proof<TargetField, BaseField, PC>,
     ) -> Result<bool, MarlinError> {
         let verifier_time = start_timer!(|| "Marlin::Verify");
+        let first_commitments = &proof.commitments[0];
+        let second_commitments = &proof.commitments[1];
+        let third_commitments = &proof.commitments[2];
+        let proof_has_correct_zk_mode = if MM::ZK {
+            first_commitments.len() == 4 && proof.pc_proof.is_hiding()
+        } else {
+            first_commitments.len() == 3 && !proof.pc_proof.is_hiding()
+        };
+        if !proof_has_correct_zk_mode {
+            eprintln!(
+                "Too many commitments in the first round ({}) or proof has incorrect hiding mode ({})",
+                first_commitments.len(),
+                proof.pc_proof.is_hiding()
+            );
+            return Ok(false);
+        }
 
-        let public_input = {
+        let padded_public_input = {
             let domain_x = EvaluationDomain::<TargetField>::new(public_input.len() + 1).unwrap();
 
             if cfg!(debug_assertions) {
@@ -584,39 +613,35 @@ where
                 println!("Size of evaluation domain x: {}", domain_x.size());
             }
 
-            let mut new_input = public_input.to_vec();
-            new_input.resize(
-                core::cmp::max(public_input.len(), domain_x.size() - 1),
-                TargetField::zero(),
-            );
+            let mut new_input = vec![TargetField::one()];
+            new_input.extend_from_slice(public_input);
+            new_input.resize(core::cmp::max(public_input.len(), domain_x.size()), TargetField::zero());
+            assert!(new_input.first().unwrap().is_one());
             new_input
         };
+        let public_input = ProverConstraintSystem::unformat_public_input(&padded_public_input);
 
         if cfg!(debug_assertions) {
-            println!("Number of padded public variables: {}", public_input.len());
+            println!("Number of padded public variables: {}", padded_public_input.len());
         }
 
-        let is_recursion = MM::RECURSION;
+        let mut fs_rng = FS::with_parameters(fs_parameters);
 
-        let mut fs_rng = FS::new();
-
-        if is_recursion {
+        if MM::RECURSION {
             fs_rng.absorb_bytes(&to_bytes_le![&Self::PROTOCOL_NAME].unwrap());
-            fs_rng.absorb_native_field_elements(&compute_vk_hash::<TargetField, BaseField, PC, FS>(
-                circuit_verifying_key,
-            )?);
-            fs_rng.absorb_nonnative_field_elements(&public_input, OptimizationType::Weight);
+            fs_rng.absorb_native_field_elements(&circuit_verifying_key.circuit_commitments);
+            fs_rng.absorb_nonnative_field_elements(&padded_public_input, OptimizationType::Weight);
         } else {
-            fs_rng.absorb_bytes(&to_bytes_le![&Self::PROTOCOL_NAME, &circuit_verifying_key, &public_input].unwrap());
+            fs_rng.absorb_bytes(
+                &to_bytes_le![&Self::PROTOCOL_NAME, &circuit_verifying_key, padded_public_input].unwrap(),
+            );
         }
 
         // --------------------------------------------------------------------
         // First round
 
-        let first_commitments = &proof.commitments[0];
-
-        if is_recursion {
-            fs_rng.absorb_native_field_elements(&first_commitments);
+        if MM::RECURSION {
+            fs_rng.absorb_native_field_elements(first_commitments);
             if !proof.prover_messages[0].field_elements.is_empty() {
                 fs_rng.absorb_nonnative_field_elements(
                     &proof.prover_messages[0].field_elements,
@@ -627,15 +652,14 @@ where
             fs_rng.absorb_bytes(&to_bytes_le![first_commitments, proof.prover_messages[0]].unwrap());
         }
 
-        let (_, verifier_state) = AHPForR1CS::verifier_first_round(circuit_verifying_key.circuit_info, &mut fs_rng)?;
+        let (_, verifier_state) =
+            AHPForR1CS::<_, MM>::verifier_first_round(circuit_verifying_key.circuit_info, &mut fs_rng)?;
         // --------------------------------------------------------------------
 
         // --------------------------------------------------------------------
         // Second round
-        let second_commitments = &proof.commitments[1];
-
-        if is_recursion {
-            fs_rng.absorb_native_field_elements(&second_commitments);
+        if MM::RECURSION {
+            fs_rng.absorb_native_field_elements(second_commitments);
             if !proof.prover_messages[1].field_elements.is_empty() {
                 fs_rng.absorb_nonnative_field_elements(
                     &proof.prover_messages[1].field_elements,
@@ -646,15 +670,13 @@ where
             fs_rng.absorb_bytes(&to_bytes_le![second_commitments, proof.prover_messages[1]].unwrap());
         }
 
-        let (_, verifier_state) = AHPForR1CS::verifier_second_round(verifier_state, &mut fs_rng)?;
+        let (_, verifier_state) = AHPForR1CS::<_, MM>::verifier_second_round(verifier_state, &mut fs_rng)?;
         // --------------------------------------------------------------------
 
         // --------------------------------------------------------------------
         // Third round
-        let third_commitments = &proof.commitments[2];
-
-        if is_recursion {
-            fs_rng.absorb_native_field_elements(&third_commitments);
+        if MM::RECURSION {
+            fs_rng.absorb_native_field_elements(third_commitments);
             if !proof.prover_messages[2].field_elements.is_empty() {
                 fs_rng.absorb_nonnative_field_elements(
                     &proof.prover_messages[2].field_elements,
@@ -665,7 +687,7 @@ where
             fs_rng.absorb_bytes(&to_bytes_le![third_commitments, proof.prover_messages[2]].unwrap());
         }
 
-        let verifier_state = AHPForR1CS::verifier_third_round(verifier_state, &mut fs_rng)?;
+        let verifier_state = AHPForR1CS::<_, MM>::verifier_third_round(verifier_state, &mut fs_rng)?;
         // --------------------------------------------------------------------
 
         // Collect degree bounds for commitments. Indexed polynomials have *no*
@@ -674,15 +696,11 @@ where
         let index_info = circuit_verifying_key.circuit_info;
         let degree_bounds = vec![None; circuit_verifying_key.circuit_commitments.len()]
             .into_iter()
-            .chain(AHPForR1CS::prover_first_round_degree_bounds(&index_info))
-            .chain(AHPForR1CS::prover_second_round_degree_bounds(&index_info))
-            .chain(AHPForR1CS::prover_third_round_degree_bounds(&index_info));
+            .chain(AHPForR1CS::<_, MM>::prover_first_round_degree_bounds(&index_info))
+            .chain(AHPForR1CS::<_, MM>::prover_second_round_degree_bounds(&index_info))
+            .chain(AHPForR1CS::<_, MM>::prover_third_round_degree_bounds(&index_info));
 
-        let polynomial_labels: Vec<String> = if is_recursion {
-            AHPForR1CS::<TargetField>::polynomial_labels_with_vanishing().collect()
-        } else {
-            AHPForR1CS::<TargetField>::polynomial_labels().collect()
-        };
+        let polynomial_labels = AHPForR1CS::<TargetField, MM>::polynomial_labels();
 
         // Gather commitments in one vector.
         let commitments: Vec<_> = circuit_verifying_key
@@ -696,9 +714,9 @@ where
             .map(|((c, l), d)| LabeledCommitment::new(l, c, d))
             .collect();
 
-        let (query_set, verifier_state) = AHPForR1CS::verifier_query_set(verifier_state, &mut fs_rng, is_recursion);
+        let (query_set, verifier_state) = AHPForR1CS::<_, MM>::verifier_query_set(verifier_state, &mut fs_rng);
 
-        if is_recursion {
+        if MM::RECURSION {
             fs_rng.absorb_nonnative_field_elements(&proof.evaluations, OptimizationType::Weight);
         } else {
             fs_rng.absorb_bytes(&to_bytes_le![&proof.evaluations].unwrap());
@@ -708,11 +726,11 @@ where
 
         let mut evaluation_labels = Vec::<(String, TargetField)>::new();
 
-        for q in query_set.iter().cloned() {
-            if AHPForR1CS::<TargetField>::LC_WITH_ZERO_EVAL.contains(&q.0.as_ref()) {
-                evaluations.insert(q, TargetField::zero());
+        for (label, (_point_name, q)) in query_set.iter().cloned() {
+            if AHPForR1CS::<TargetField, MM>::LC_WITH_ZERO_EVAL.contains(&label.as_ref()) {
+                evaluations.insert((label, q), TargetField::zero());
             } else {
-                evaluation_labels.push(q);
+                evaluation_labels.push((label, q));
             }
         }
         evaluation_labels.sort_by(|a, b| a.0.cmp(&b.0));
@@ -720,10 +738,9 @@ where
             evaluations.insert(q, *eval);
         }
 
-        let lc_s =
-            AHPForR1CS::construct_linear_combinations(&public_input, &evaluations, &verifier_state, is_recursion)?;
+        let lc_s = AHPForR1CS::<_, MM>::construct_linear_combinations(&public_input, &evaluations, &verifier_state)?;
 
-        let evaluations_are_correct = if is_recursion {
+        let evaluations_are_correct = if MM::RECURSION {
             let num_open_challenges: usize = 7;
 
             let mut opening_challenges = Vec::new();
@@ -757,22 +774,34 @@ where
         };
 
         if !evaluations_are_correct {
+            #[cfg(debug_assertions)]
             eprintln!("PC::Check failed");
         }
         end_timer!(verifier_time, || format!(
             " PC::Check for AHP Verifier linear equations: {}",
-            evaluations_are_correct
+            evaluations_are_correct & proof_has_correct_zk_mode
         ));
-        Ok(evaluations_are_correct)
+        Ok(evaluations_are_correct & proof_has_correct_zk_mode)
     }
 
     /// Verify that a proof for the constraint system defined by `C` asserts that
     /// all constraints are satisfied using the prepared verifying key.
     pub fn prepared_verify(
-        prepared_vk: &PreparedCircuitVerifyingKey<TargetField, PC>,
+        prepared_vk: &PreparedCircuitVerifyingKey<TargetField, BaseField, PC, MM>,
         public_input: &[TargetField],
-        proof: &Proof<TargetField, PC>,
+        proof: &Proof<TargetField, BaseField, PC>,
     ) -> Result<bool, MarlinError> {
         Self::verify(&prepared_vk.orig_vk, public_input, proof)
     }
+
+    /// Verify that a proof for the constraint system defined by `C` asserts that
+    /// all constraints are satisfied using the prepared verifying key.
+    pub fn prepared_verify_with_fs_parameters(
+        prepared_vk: &PreparedCircuitVerifyingKey<TargetField, BaseField, PC, MM>,
+        fs_parameters: &FS::Parameters,
+        public_input: &[TargetField],
+        proof: &Proof<TargetField, BaseField, PC>,
+    ) -> Result<bool, MarlinError> {
+        Self::verify_with_fs_parameters(&prepared_vk.orig_vk, fs_parameters, public_input, proof)
+    }
 }
